/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_INCREMENTALASSIMILATION_H_
#define OOPS_ASSIMILATION_INCREMENTALASSIMILATION_H_

#include <memory> // for std::shared_ptr, std::unique_ptr
#include <vector>

#include "eckit/config/Configuration.h"
#include "eckit/memory/NonCopyable.h"

#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/Minimizer.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/PostProcessorTL.h"
#include "oops/base/StateInfo.h"
#include "oops/base/StateSaver.h"
#include "oops/base/StateWriter.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"

#include "util/Logger.h"
#include "util/LogbookWriter.h"
#include "util/Restart.h"

namespace oops {

/*
 * Note: two flavours of incremental assimilations are implemented:
 *         A. dx = U dv
 *         B. dx = U dv + x_b - x_fg
 *       The later is not yet compatible with the restart mechanism, when
 *         the late 4D start mechanism is activated
 */

template<typename MODEL>
class IncrementalAssimilation : private eckit::NonCopyable {
  typedef ControlIncrement<MODEL>    CtrlInc_;
  typedef ControlVariable<MODEL>     CtrlVariable_;
  typedef Increment<MODEL>           Increment_;
  typedef Minimizer<MODEL>           Minimizer_;
  typedef State<MODEL>               State_;
  typedef Variables<MODEL>           Variables_;
 public:
  static const std::string classname() {return "oops::IncrementalAssimilation";}

/// Constructor, destructor
  IncrementalAssimilation(CostFunction<MODEL> &);
  ~IncrementalAssimilation() {}

/// Perform assimilation
  void assimilate(const eckit::Configuration &);
  std::shared_ptr<CtrlVariable_> getAnalysis() const {ASSERT(xx_); return xx_;}
  const bool & isLate4DstartCompleted() const {return late4Dcompleted_;}

 private:
/// Restart mechanism utilities
  void checkpoint(eckit::LocalConfiguration & conf) const;
  void restart(const eckit::Configuration & conf) const;
  void connectToRestart(util::Restart & restarter) const;

/// Late 4D mechanims utilities
  void late4Dstart(const eckit::LocalConfiguration &,
                   const std::shared_ptr<StateSaver<State_> >,
                   const CtrlInc_ &);

  const bool & isLate4D() const {return late4D_;}
  void setLate4D(const bool & active) {late4D_ = active;}
  void setLate4Dcompleted(const bool & active) {late4Dcompleted_ = active;}
  const bool isLate4DstartActive(const size_t & nouter) const;

/// Data members
  CostFunction<MODEL> & J_;
  std::shared_ptr<ControlVariable<MODEL> > xx_;
  std::unique_ptr<Minimizer<MODEL> > minim_;

  size_t jouter_;
  mutable size_t jstart_;
  bool late4D_;
  bool late4Dcompleted_;
};

// =============================================================================

template<typename MODEL>
IncrementalAssimilation<MODEL>::IncrementalAssimilation(CostFunction<MODEL> & J)
  : J_(J), xx_(), minim_(), jouter_(0), jstart_(0),
    late4D_(false), late4Dcompleted_(false)
{
  Log::trace() << "IncrementalAssimilation:constructed" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrementalAssimilation<MODEL>::assimilate(const eckit::Configuration & config) {

//  Initialize first guess from background
  xx_.reset(new CtrlVariable_(J_.jb().getBackground()));
  Log::trace() << "IncrementalAssimilation: first guess has been set up" << std::endl;

// Setup outer loop
  std::vector<eckit::LocalConfiguration> iterconfs;
  config.get("variational.iteration", iterconfs);
  size_t nouter = iterconfs.size();
  Log::info() << "Running incremental assimilation with " << nouter
              << " outer iterations." << std::endl;

// Setup minimizer
  eckit::LocalConfiguration minConf(config, "minimizer");
  const size_t nnout = nouter;
  minConf.set("nouter", nnout);
  minim_.reset(MinFactory<MODEL>::create(minConf, J_));

// Late4D
  std::shared_ptr< StateSaver<State_> > stateSaver;
  std::unique_ptr<eckit::LocalConfiguration> late4Dconf;
  if (config.has("late4d")) {
    this->setLate4D(true);
    late4Dconf.reset(new eckit::LocalConfiguration(config, "late4d"));
  }

// Restart mechanism
  util::Restart restarter({"FirstGuess", "Minimizer", "IncrementalAssimilation"});
  this->connectToRestart(restarter);
  restarter.restart({"FirstGuess", "IncrementalAssimilation"});

// Warm start first guess
  if (not restarter.isRestarted() &&
      config.has("variational.WarmStart.FirstGuess")) {
    xx_->read(config.getSubConfiguration("variational.WarmStart.FirstGuess"));
  }

  for (jouter_ = jstart_; jouter_ < nouter; ++jouter_) {
    util::LogbookWriter<size_t> log("OuterLoop", jouter_);

//  Get configuration for current outer iteration
    Log::info() << "IncrementalAssimilation: Configuration for outer iteration "
                << jouter_ << ":\n" << iterconfs[jouter_];

//  Setup for the trajectory run
    PostProcessor<State_> post;
    // Setup diagnostic prints
    if (iterconfs[jouter_].has("prints")) {
      const eckit::LocalConfiguration prtConfig(iterconfs[jouter_], "prints");
      post.enrollProcessor(new StateInfo<State_>("traj", prtConfig));
    }
    // Setup state saver for Late4D
    if (this->isLate4DstartActive(nouter)) {
      const eckit::LocalConfiguration late4DstateConf(*late4Dconf, "state");
      stateSaver.reset(new StateSaver<State_>(late4DstateConf));
      post.enrollProcessor(stateSaver);
    }

//  Setup quadratic problem
    J_.linearize(*xx_, iterconfs[jouter_], post);

// Warm start minimizer
    if (jouter_ == 0 && config.has("variational.WarmStart.Minimizer")) {
      minim_->warmStart(config.getSubConfiguration("variational.WarmStart.Minimizer"));
    }

//  Restart minimizer
    if (jouter_ == jstart_)
      restarter.restart({"Minimizer"});

//  Minimization
    std::unique_ptr<CtrlInc_> dx(minim_->minimize(iterconfs[jouter_]));

//  Add increment
    J_.addIncrement(*xx_, *dx);

//  Write increment
    if (iterconfs[jouter_].has("increment"))
      dx->write(iterconfs[jouter_].getSubConfiguration("increment"));

//  Compute and save 4D analysis/increment or save increment
    if (this->isLate4DstartActive(nouter)) {
      // Calculate and write an fields, write increment
      this->late4Dstart(*late4Dconf, stateSaver, *dx);
      // Clean-up
      stateSaver.reset();
    }

//  Clean-up trajectory, etc...
    J_.resetLinearization();

//  Checkpointing
    restarter.checkpoint(iterconfs[jouter_], {"FirstGuess", "CostFunction",
                                              "Minimizer", "IncrementalAssimilation"});
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
const bool
IncrementalAssimilation<MODEL>::isLate4DstartActive(const size_t & nouter) const {
  bool late4Dactive = false;
  /*
   * Active late 4D start mechanism in the last outer loop if enabled through config
   */
  if (this->isLate4D() && jouter_ == nouter - 1)
    late4Dactive = true;
  return late4Dactive;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrementalAssimilation<MODEL>::checkpoint(eckit::LocalConfiguration & conf) const {
  conf.set("Iteration", jouter_);
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrementalAssimilation<MODEL>::restart(const eckit::Configuration & conf) const {
  jstart_ = conf.getUnsigned("Iteration");
  ++jstart_;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrementalAssimilation<MODEL>::connectToRestart(util::Restart & restarter) const {
// Checkpoint connect elements
  restarter.connectChkptCallback("CostFunction",
                                 [this](eckit::LocalConfiguration & conf)
                                 { return J_.checkpoint(conf); });
  restarter.connectChkptCallback("Minimizer",
                                 [this](eckit::LocalConfiguration & conf)
                                 { return minim_->checkpoint(conf); });
  restarter.connectChkptCallback("FirstGuess",
                                 [this](eckit::LocalConfiguration & conf)
                                 { return xx_->write(conf); });
  restarter.connectChkptCallback("IncrementalAssimilation",
                                 [this](eckit::LocalConfiguration & conf)
                                 { return this->checkpoint(conf); });

// Restart connect elements
  restarter.connectRstrtCallback("CostFunction",
                                 [this](const eckit::Configuration & conf)
                                 { return J_.restart(conf); });
  restarter.connectRstrtCallback("Minimizer",
                                 [this](const eckit::Configuration & conf)
                                 { return minim_->restart(conf); });
  restarter.connectRstrtCallback("FirstGuess",
                                 [this](const eckit::Configuration & conf)
                                 { return xx_->read(conf); });
  restarter.connectRstrtCallback("IncrementalAssimilation",
                                 [this](const eckit::Configuration & conf)
                                 { return this->restart(conf); });
}

// -----------------------------------------------------------------------------

/*
 * Late 4D start mechanism can be applied in the last outer loop.
 *
 *  The mechanims consists of the following steps:
 *   A.) Record x_fg^(steps) during the first guess trajectory;
 *   B.) Propagate the tangent linear model and record dx_fg^(steps);
 *   C.) Perform sanity checks on dx_fg^(steps);
 *   D.) Obtain 'an' fields: x_an^(steps) = x_fg^(steps) + dx_fg^(steps);
 *   E.) Write out 'an' fields and dx_fg^(steps);
 */

template<typename MODEL>
void IncrementalAssimilation<MODEL>::late4Dstart(const eckit::LocalConfiguration & conf,
                                const std::shared_ptr<StateSaver<State_> > stateSaver,
                                const CtrlInc_ & dx) {
  Log::info() << "IncrementalAssimilation::late4Dstart starting" << std::endl;

  const eckit::LocalConfiguration late4DstateConf(conf, "state");

  // Propagate increment to requested times
  PostProcessorTL<Increment_> cost;
  PostProcessor<Increment_> post;

  std::shared_ptr< StateSaver<Increment_> > incrementSaver;
  incrementSaver.reset(new StateSaver<Increment_>(late4DstateConf));
  post.enrollProcessor(incrementSaver);

  CtrlInc_ dxtmp(dx);
  J_.runTLM(dxtmp, cost, post);

  // Retrieve increments
  const std::vector<Increment_> & increments(incrementSaver->releaseFields());

  // Increment sanity checks
  bool lSanityCheckPassed = true;
  std::vector<eckit::LocalConfiguration> confs;
  if (late4DstateConf.has("incrementSanityChecks"))
    late4DstateConf.get("incrementSanityChecks", confs);

  for (auto const & config: confs) {
    const double zmin = config.getDouble("min");
    const double zmax = config.getDouble("max");
    Variables_ var(config);
    for (auto const & incr: increments) {
      Log::info() << "IncrementalAssimilation::late4Dstart " << std::endl
                  << "  Sanity Test: variable " << var << std::endl
                  << "  Sanity Test: bounds (" << zmin << ", " << zmax << ") " << std::endl
                  << "  Sanity Test: increment valid at: " << incr.validTime() << std::endl;
      const double zincmin = incr.min(var);
      const double zincmax = incr.max(var);
      Log::info() << "  Sanity Test: zmax = " << zincmax << std::endl
                  << "  Sanity Test: zmin = " << zincmin << std::endl;
      if (zincmin >= zmin && zincmax <= zmax) {
        Log::info() << "  Sanity Test: passed" << std::endl;
      } else {
        Log::info() << "  Sanity Test: NOT passed" << std::endl;
        lSanityCheckPassed = false;
        // Don't skip final trajectory
        this->setLate4Dcompleted(false);
      }
    }
  }

  if (lSanityCheckPassed) {
    Log::info() << "IncrementalAssimilation::late4Dstart " << std::endl
                << "  All increment sanity tests PASSED   " << std::endl
                << "  Skipping final non-linear trajectory" << std::endl;

    // Write ObsAux and ModAux fields (+0H)
    xx_->obsVar().write(conf);
    xx_->modVar().write(conf);

    // Close ODB
    J_.resetLinearization();

    // Retrieve saved FG states
    std::vector<State_> & states(stateSaver->releaseFields());

    // Add increments to FG states and write
    ASSERT(states.size() == increments.size());
    for (size_t jj = 0; jj < states.size(); ++jj) {
      states[jj] += increments[jj];
      // Write "an" fields
      states[jj].write(late4DstateConf);
      // Write "dx_fg" fields except at the begining of the window (handled outside)
      if (increments[jj].validTime() != xx_->state()[0].validTime())
        increments[jj].write(late4DstateConf);
      Log::info() << "IncrementalAssimilation::late4Dstart " << std::endl
                  << "  State An valid at: " << states[jj].validTime()
                  << " written out." << std::endl;
    }

    // Skip final trajectory
    this->setLate4Dcompleted(true);
  }

  Log::info() << "IncrementalAssimilation::late4Dstart finished" << std::endl;
}

// =============================================================================

}  // namespace oops

#endif  // OOPS_ASSIMILATION_INCREMENTALASSIMILATION_H_
