/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_SQRTMINIMIZER_H_
#define OOPS_ASSIMILATION_SQRTMINIMIZER_H_

#include <iterator>
#include <map>
#include <memory> // for std::unique_ptr
#include <string>
#include <vector>

#include "eckit/config/Configuration.h"

#include "oops/assimilation/BMatrix.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVector.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/JbMatrix.h"
#include "oops/assimilation/UtHtRinvHdxFG.h"
#include "oops/assimilation/UtHtRinvHUMatrix.h"
#include "oops/assimilation/Minimizer.h"

#include "util/dot_product.h"
#include "util/formats.h"
#include "util/Logger.h"

namespace oops {

/// SQRT Minimizers
/*!
 * SQRTMinimizer is the base class for all minimizers that use \f$ U\f$ to
 * precondition the variational minimisation problem and use the auxiliary
 * variable \f$ dx = U \hat{v} \f$ or \f$ dx = U \hat{v} + x_b - x_fg \f$
 */

// -----------------------------------------------------------------------------

template<typename MODEL> class SQRTMinimizer : public Minimizer<MODEL> {
  typedef BMatrix<MODEL>           Bmat_;
  typedef CostFunction<MODEL>      CostFct_;
  typedef ControlIncrement<MODEL>  CtrlInc_;
  typedef ControlVector<MODEL>     CtrlVec_;
  typedef JbMatrix<MODEL>          Jbmat_;
  typedef UtHtRinvHUMatrix<MODEL>  UtHtRinvHU_;
  typedef UtHtRinvHdxFG<MODEL>     UtHtRinvHdxFG_;
  typedef Minimizer<MODEL>         Minimizer_;

 public:
  virtual const std::string classname() const override =0;

/// Constructor, destructor
  SQRTMinimizer(const eckit::Configuration & conf, const CostFct_ & J)
   : Minimizer_(J), J_(J),
     lclassicChaVar_(true),
     linfoConv_(false), lrunOnlineAdjTest_(false),
     ninner_(30), ninnerMin_(0), gnreduc_(1.e-14),
     gradJb_(), dv0_(), dvh_(),
     costJ0Jb_(0), costJ0JbJq_(0),
     setObsAuxFormulation_(false), obsAuxFormulation_(),
     iter_(0), maxiter_(999)
  {
     this->setupSolver(conf);
  }
  ~SQRTMinimizer() {}

/// Checkpointing, restart
  void checkpoint(eckit::LocalConfiguration &) const;
  void restart(const eckit::Configuration &);
  void warmStart(const eckit::Configuration &);

 private:
  CtrlInc_ * doMinimize(const eckit::Configuration &) override;
  virtual double solve(CtrlVec_ &, CtrlVec_ &,
                       const Jbmat_ &,
                       const UtHtRinvHU_ &,
                       const bool &,
                       const bool &,
                       const size_t &,
                       const size_t &,
                       const double &,
                       const double &,
                       const double &) =0;

/// Checkpointing, restart of the limited memory preconditioner
  virtual void checkpointLMP(eckit::LocalConfiguration &) const =0;
  virtual void restartLMP(const eckit::Configuration &) =0;

  void checkpointGradJb(eckit::LocalConfiguration &) const;
  void checkpointDv0(eckit::LocalConfiguration &) const;
  void checkpointDvh(eckit::LocalConfiguration &) const;
  void checkpointAux(eckit::LocalConfiguration &) const;

  void restartGradJb(const eckit::Configuration &);
  void restartDv0(const eckit::Configuration &);
  void restartDvh(const eckit::Configuration &);
  void restartAux(const  eckit::Configuration &);

/// Quadratic cost function calculation (Jb component)
  void updateJbCostFunction(const CtrlVec_ & dv, const CtrlVec_ & jbdv);

/// Setup controls
  void setupSolver(const eckit::Configuration &);
  void setupMinimization(const eckit::Configuration &);

/// ObsAux formulation
  void enforceObsAuxFormulation(CtrlVec_ & dv) const;

/// Data members
  const CostFct_ & J_;

/// Controls
  bool lclassicChaVar_;
  bool linfoConv_;
  bool lrunOnlineAdjTest_;
  size_t ninner_;
  size_t ninnerMin_;
  double gnreduc_;

/// Control vectors
  std::unique_ptr<CtrlVec_> gradJb_;
  std::unique_ptr<CtrlVec_> dv0_;
  std::vector<CtrlVec_> dvh_;

// Cost function Jb
  double costJ0Jb_;
  double costJ0JbJq_;

// ObsAux formulation
  mutable bool setObsAuxFormulation_;
  mutable std::map<std::string, bool> obsAuxFormulation_;

 protected:
  size_t iter_;
  mutable size_t maxiter_;
};

// =============================================================================

template<typename MODEL>
ControlIncrement<MODEL> * SQRTMinimizer<MODEL>::doMinimize(const eckit::Configuration & config) {
  iter_++;
  ASSERT(iter_ <= maxiter_);
  Log::info() << std::endl;
  Log::info() << classname() << ": starting" << std::endl;

// Setup minimization specific controls
  this->setupMinimization(config);

// Define the matrices
  const Jbmat_      Jb(J_);
  const Bmat_       B(J_);
  const UtHtRinvHU_ UtHtRinvHU(J_, lrunOnlineAdjTest_);

// Compute U^T H^T R^{-1} d
  CtrlInc_ fgrhs(J_.jb().getFirstGuessIncrement(), false);
  J_.computeGradientFG(fgrhs);
  CtrlVec_ rhs(J_.jb());
  B.multiplySqrtTrans(fgrhs, rhs);

// Create a contorl vector
  CtrlVec_ dv(J_.jb());

  if (lclassicChaVar_) {
    // dx = U dv

    // Interpolate gradJb to the new inner loop resolution
    if (gradJb_ == 0) {
      gradJb_.reset(new CtrlVec_(J_.jb()));
    } else {
      gradJb_.reset(new CtrlVec_(J_.jb(), *gradJb_));
    }
    // Calculate first guess gradient
    // -SUM dv_i + U^T H^T R^{-1} d
    J_.jb().addGradientFG(rhs, *gradJb_);
  } else {
    // dx = U dv + x_b - x_fg

    // Interpolate initial guess to the new inner loop resolution
    if (dv0_ == 0) {
      dv0_.reset(new CtrlVec_(J_.jb()));
    } else {
      dv0_.reset(new CtrlVec_(J_.jb(), *dv0_));
    }
    // Calculate first guess gradient
    // U^T H^T R^{-1} [d + H (x_fg - x_b)]
    const UtHtRinvHdxFG_ UtHtRinvHdxFG(J_);
    CtrlVec_ ut_ht_rinv_h_dxfg(J_.jb());
    UtHtRinvHdxFG.compute(ut_ht_rinv_h_dxfg);
    rhs += ut_ht_rinv_h_dxfg;
    // initialise dv_k to dv_{k-1}
    dv = *dv0_;
  }

// Set J[0] = 0.5 (x_i - x_b)^T B^{-1} (x_i - x_b) + 0.5 d^T R^{-1} d
  const double costJ0JoJc = J_.getCostJoJc();

  Log::info() << "DEBUG Jq: before minimization = "
              << util::full_precision(dot_product(dv.modVar(), dv.modVar()))
              << std::endl;

// Solve the linear system
  double reduc = this->solve(dv, rhs, Jb, UtHtRinvHU,
                             lclassicChaVar_, linfoConv_,
                             ninner_, ninnerMin_, gnreduc_,
                             costJ0Jb_, costJ0JoJc);

  Log::info() << "DEBUG Jq: after minimization = "
              << util::full_precision(dot_product(dv.modVar(), dv.modVar()))
              << std::endl;

// Set ObsAux componenet of the control vector
  this->enforceObsAuxFormulation(dv);

// Calculate the increment dx_k = U dv
  CtrlInc_ * dx = new CtrlInc_(J_.jb().getFirstGuessIncrement(), false);
  B.multiplySqrt(dv, *dx);

// Print
  if (linfoConv_) {
    Log::test() << classname() << ": final relative gain in information content = "
                               << reduc << std::endl;
  } else {
    Log::test() << classname() << ": reduction in residual norm = "
                               << reduc << std::endl;
  }
  Log::info() << classname() << " output increment:" << *dx << std::endl;

// Calculate quadratic cost function
  CtrlVec_ jbdv(dv, false);
  Jb.multiply(dv, jbdv);

// Update Jb component of J[0]: 0.5 (x_i - x_b)^T B^-1 (x_i - x_b)
  this->updateJbCostFunction(dv, jbdv);

// Book-keeping for next minimization/(transformation of the increment)
  if (lclassicChaVar_) {
    // dx = U dv

    // Update gradient Jb
    *gradJb_ += dv;
  } else {
    // dx = U dv + x_b - x_fg

    // Save inital guess
    *dv0_ = dv;
    // Tranform dx_k to be defined w.r.t. the first guess
    const CtrlInc_ & dxfg = J_.jb().getFirstGuessIncrement();
    *dx -= dxfg;
  }

  Log::info() << classname() << ": done" << std::endl;

  return dx;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::setupSolver(const eckit::Configuration & conf) {
  // Change of variable
  if (conf.has("classic_change_of_variables")) {
    // true  : dx = U dv
    // false : dx = U dv + x_b - x_fg
    lclassicChaVar_ = conf.getBool("classic_change_of_variables");
  }

  // Check if infromation content based convergence criteria enabled
  if (conf.has("linfoconv")) linfoConv_ = conf.getBool("linfoconv");

  // Check if LMP preconditioner is enabled
  if (conf.has("preconditioner") || conf.has("obsAuxFormulations")) {
    // Provide number of outer loops in order to skip the update of the
    // preconditioner in the final minimization or/and to properly apply
    // obs aux incremental formulation
    ASSERT(conf.has("minimizations"));
    maxiter_ = conf.getUnsigned("minimizations");
  }

  // Set ObsAux formulation (incremental/non-incremental)
  if (conf.has("obsAuxFormulations")) {
    Log::info() << "SQRTMinimizer<MODEL>::setupSolver() set ObsAux* formulations"
                << std::endl;
    setObsAuxFormulation_ = true;
    std::vector<eckit::LocalConfiguration> configs;
    conf.get("obsAuxFormulations", configs);
    for (size_t ii = 0; ii < configs.size(); ++ii) {
       std::string key(configs[ii].getString("component"));
       bool value(configs[ii].getBool("incremental"));
       obsAuxFormulation_.insert(std::make_pair(key, value));
       Log::info() << "   ObsAuxComponent: " << key
                   << ", incremental formulation: " << value
                   << std::endl;
    }
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::setupMinimization(const eckit::Configuration & conf) {
  // Set the number of iterations
  if (conf.has("ninner"))
    ninner_ = conf.getUnsigned("ninner");
  ASSERT(ninner_ > ninnerMin_);

  // Set the tolerance
  if (conf.has("gradient_norm_reduction"))
    gnreduc_ = conf.getDouble("gradient_norm_reduction");

  // Activate online adjoint tests
  if (conf.has("onlineDiagnostics")) {
    const eckit::LocalConfiguration onlineDiag(conf, "onlineDiagnostics");
    lrunOnlineAdjTest_ = onlineDiag.getBool("onlineAdjTest");
  } else {
    lrunOnlineAdjTest_ = false;
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::enforceObsAuxFormulation(CtrlVec_ & dv) const {
   if (setObsAuxFormulation_ && iter_ < maxiter_) {
      Log::info() << "SQRTMinimizer<MODEL>::enforceObsAuxFormulation(dv)" << std::endl;
      std::map<std::string, bool>::const_iterator icst = obsAuxFormulation_.begin();
      while (icst != obsAuxFormulation_.end())
      {
         if (!icst->second) {
           for (std::size_t jobs = 0; jobs < dv.obsVar().size(); ++jobs) {
             if ( dv.obsVar()[jobs].size() > 0) {
               dv.obsVar()[jobs].obsAuxCtlVec(icst->first).zero();
               Log::info() << "  set component of the control vector: "
                           << icst->first << " to zero" << std::endl;
             }
           }
         }
         ++icst;
      }
   }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::updateJbCostFunction(const CtrlVec_ & dv,
                                                const CtrlVec_ & jbdv) {
  // Store jbdvs
  if (lclassicChaVar_) {
    dvh_.push_back(jbdv);

    // Update Jb component of J[0]: 0.5 (x_i - x_b)^T B^-1 (x_i - x_b)
    costJ0Jb_ += 0.5 * dot_product(dv, jbdv);
    costJ0JbJq_ += 0.5 * dot_product(dv.modVar(), jbdv.modVar());

    for (size_t jouter = 1; jouter < dvh_.size(); ++jouter) {
      CtrlVec_ dvhtmp(J_.jb(), dvh_[jouter-1]);
      // MC temp
      //costJ0Jb_ += dot_product(dv.state(), dvhtmp.state());
      costJ0JbJq_ += dot_product(dv.modVar(), dvhtmp.modVar());
    }
  } else {
    // Update Jb component of J[0]: 0.5 (x_i - x_b)^T B^-1 (x_i - x_b)
    costJ0Jb_ = 0.5 * dot_product(dv, jbdv);
    costJ0JbJq_ = 0.5 * dot_product(dv.modVar(), jbdv.modVar());
  }

  // Print out costJ0Jb and its ObsAux components
  Log::info() << "SQRTMinimizer: costJ0Jb      : "
              << util::full_precision(costJ0Jb_)
              << std::endl;
  Log::info() << "SQRTMinimizer: costJ0JbJq    : "
              << util::full_precision(costJ0JbJq_)
              << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::checkpoint(eckit::LocalConfiguration & conf) const {
  Log::info() << " " << classname() << "::checkpoint starting" << std::endl;

  // GradJb
  checkpointGradJb(conf);

  // dv0
  checkpointDv0(conf);

  // dvh
  checkpointDvh(conf);

  // costJ0Jb
  checkpointAux(conf);

  // LMP
  if (conf.has("preconditioner"))
    this->checkpointLMP(conf);

  // iteration
  conf.set("outerIteration", iter_);

  Log::info() << " " << classname() << "::checkpoint done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::warmStart(const eckit::Configuration & conf) {
  Log::info() << " " << classname() << "::warmStart starting" << std::endl;

  // GradJb
  if (conf.has("gradJb"))
    restartGradJb(conf);

  // dv0
  if (conf.has("dv0"))
    restartDv0(conf);

  // dvh
  if (conf.has("dvh"))
    restartDvh(conf);

  // costJ0Jb
  if (conf.has("quadraticCost"))
    restartAux(conf.getSubConfiguration("quadraticCost"));

  // LMP
  if (conf.has("preconditioner"))
    this->restartLMP(conf);

  Log::info() << " " << classname() << "::warmStart done" << std::endl;
}

// -----------------------------------------------------------------------------


template<typename MODEL>
void SQRTMinimizer<MODEL>::restart(const eckit::Configuration & conf) {
  Log::info() << " " << classname() << "::restart starting" << std::endl;

  // GradJb
  restartGradJb(conf);

  // dv0
  restartDv0(conf);

  // dvh
  restartDvh(conf);

  // costJ0Jb
  restartAux(conf);

  // LMP
  if (conf.has("preconditioner")) {
    this->restartLMP(conf);
  }

  // iteration
  iter_ = conf.getUnsigned("outerIteration");

  Log::info() << " " << classname() << "::restart done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::checkpointGradJb(eckit::LocalConfiguration & conf) const {
  // GradJb
  if (lclassicChaVar_) {
    ASSERT(conf.has("gradJb"));
    {
      eckit::LocalConfiguration gradjbconf(conf, "gradJb");
      gradJb_->write(gradjbconf);
      Log::info() << " " << classname() << "::checkpoint gradJb: "
                  << util::full_precision(dot_product(*gradJb_, *gradJb_))
                  << std::endl;
    }
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::checkpointDv0(eckit::LocalConfiguration & conf) const {
  // dv0
  if (not lclassicChaVar_) {
    ASSERT(conf.has("dv0"));
    {
      eckit::LocalConfiguration dv0conf(conf, "dv0");
      ASSERT(dv0_);
      dv0_->write(dv0conf);
      Log::info() << " " << classname() << "::checkpoint dv0: "
                  << util::full_precision(dot_product(*dv0_, *dv0_))
                  << std::endl;
    }
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::checkpointDvh(eckit::LocalConfiguration & conf) const {
  // dvh
  if (lclassicChaVar_) {
    ASSERT(conf.has("dvh"));
    {
      std::vector<eckit::LocalConfiguration> dvhconfs;
      conf.get("dvh", dvhconfs);
      ASSERT(dvh_.size() == dvhconfs.size());

      for (size_t ii = 0; ii < dvh_.size(); ++ii) {
        dvh_[ii].write(dvhconfs[ii]);
        Log::info() << " " << classname() << "::checkpoint dvh[" << ii << "]: "
                    << util::full_precision(dot_product(dvh_[ii], dvh_[ii]))
                    << std::endl;
      }
    }
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::checkpointAux(eckit::LocalConfiguration & conf) const {
  // costJ0Jb
  conf.set("costJ0Jb", util::full_precision(costJ0Jb_));
  Log::info() << " " << classname() << "::checkpoint costJ0Jb: "
              << util::full_precision(costJ0Jb_) << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::restartGradJb(const eckit::Configuration & conf) {
  // GradJb
  if (lclassicChaVar_) {
    ASSERT(conf.has("gradJb"));
    eckit::LocalConfiguration gradjbconf(conf, "gradJb");
    if (gradJb_ == 0) {
      gradJb_.reset(new CtrlVec_(J_.jb()));
    }
    gradJb_->read(gradjbconf);
    Log::info() << " " << classname() << "::restart gradJb: "
                << util::full_precision(dot_product(*gradJb_, *gradJb_))
                << std::endl;
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::restartDv0(const eckit::Configuration & conf) {
  // GradJb
  if (not lclassicChaVar_) {
    ASSERT(conf.has("dv0"));
    eckit::LocalConfiguration dv0conf(conf, "dv0");
    if (dv0_ == 0) {
      dv0_.reset(new CtrlVec_(J_.jb()));
    }
    dv0_->read(dv0conf);
    Log::info() << " " << classname() << "::restart dv0: "
                << util::full_precision(dot_product(*dv0_, *dv0_))
                << std::endl;
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::restartDvh(const eckit::Configuration & conf) {
  // dvh
  if (lclassicChaVar_) {
    ASSERT(conf.has("dvh"));
    std::vector<eckit::LocalConfiguration> dvhconfs;
    conf.get("dvh", dvhconfs);
    for (size_t ii = 0; ii < dvhconfs.size(); ++ii) {
      CtrlVec_ dvh(J_.jb());
      dvh.read(dvhconfs[ii]);
      dvh_.push_back(dvh);
      Log::info() << " " << classname() << "::restart dvh[" << ii << "]: "
                  << util::full_precision(dot_product(dvh_[ii], dvh_[ii]))
                  << std::endl;
    }
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::restartAux(const eckit::Configuration & conf) {
  // costJ0Jb
  ASSERT(conf.has("costJ0Jb"));
  costJ0Jb_ = conf.getDouble("costJ0Jb");
  Log::info() << " " << classname() << "::restart costJ0Jb: "
              << util::full_precision(costJ0Jb_) << std::endl;
}

// -----------------------------------------------------------------------------



}  // namespace oops

#endif  // OOPS_ASSIMILATION_SQRTMINIMIZ_H_
