/*
 * (C) Copyright 2017-2019 UCAR
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 */

#ifndef OOPS_BASE_OBSAUXCOVARIANCES_H_
#define OOPS_BASE_OBSAUXCOVARIANCES_H_

#include <cmath>
#include <iostream>
#include <memory>
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"

#include "oops/base/ObsAuxControls.h"
#include "oops/base/ObsAuxCtlVecs.h"
#include "oops/base/ObsAuxIncrements.h"
#include "oops/base/ObsSpaces.h"
#include "oops/interface/ObsAuxCovariance.h"

#include "util/Logger.h"
#include "util/Printable.h"

namespace oops {

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxCovariances : public util::Printable,
                          private eckit::NonCopyable {
  typedef ObsAuxCovariance<MODEL>    ObsAuxCovariance_;
  typedef ObsAuxControls<MODEL>      ObsAuxControls_;
  typedef ObsAuxCtlVecs<MODEL>       ObsAuxCtlVecs_;
  typedef ObsAuxIncrements<MODEL>    ObsAuxIncrements_;
  typedef ObsSpaces<MODEL>           ObsSpaces_;

 public:
  static const std::string classname() {return "oops::ObsAuxCovariances";}

/// Constructor, destructor
  ObsAuxCovariances(const ObsSpaces_ &, const eckit::Configuration &);
  ~ObsAuxCovariances();

/// Size
  std::size_t size() const {return covs_.size();}
  const ObsAuxCovariance_ & operator[](const std::size_t ii) const {return *covs_.at(ii);}
  ObsAuxCovariance_ & operator[](const std::size_t ii) {return *covs_.at(ii);}

/// Update
  void update(const ObsAuxControls_ &, const eckit::Configuration &);

/// Operators
  void linearize(const ObsAuxControls_ &, const ObsSpaces_ &);
  void multiply(const ObsAuxIncrements_ &, ObsAuxIncrements_ &) const;
  void inverseMultiply(const ObsAuxIncrements_ &, ObsAuxIncrements_ &) const;
  void multiplySqrt(const ObsAuxCtlVecs_ &, ObsAuxIncrements_ &) const;
  void multiplySqrtTrans(const ObsAuxIncrements_ &, ObsAuxCtlVecs_ &) const;
  void randomize(ObsAuxIncrements_ &) const;

  const eckit::LocalConfiguration & config() const {return conf_;}
  const ObsSpaces_ & obspaces() const {return *odb_;}

 private:
  void print(std::ostream &) const;
  std::vector<std::shared_ptr<ObsAuxCovariance_> > covs_;
  const ObsSpaces_ mutable * odb_;
  const eckit::LocalConfiguration conf_;
};

// =============================================================================

template<typename MODEL>
ObsAuxCovariances<MODEL>::ObsAuxCovariances(const ObsSpaces_ & odb,
                                            const eckit::Configuration & conf)
  : covs_(0), odb_(), conf_(conf)
{
  Log::trace() << "ObsAuxCovariances<MODEL>::ObsAuxCovariances starting" << std::endl;
  odb_ = &odb;
  std::vector<eckit::LocalConfiguration> obsconf;
  conf.get("ObsTypes", obsconf);
  for (std::size_t jobs = 0; jobs < obsconf.size(); ++jobs) {
    std::shared_ptr<ObsAuxCovariance_>
      tmp(new ObsAuxCovariance_(odb[jobs], obsconf[jobs]));
    covs_.push_back(tmp);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::ObsAuxCovariances done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
ObsAuxCovariances<MODEL>::~ObsAuxCovariances() {
  Log::trace() << "ObsAuxCovariances<MODEL>::~ObsAuxCovariances starting" << std::endl;
  odb_ = nullptr;
  for (std::size_t jobs = 0; jobs < covs_.size(); ++jobs) {
    covs_[jobs].reset();
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::~ObsAuxCovariances done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariances<MODEL>::update(const ObsAuxControls_ & xx,
                                      const eckit::Configuration & conf) {
  Log::trace() << "ObsAuxCovariances<MODEL>::update starting" << std::endl;
  std::vector<eckit::LocalConfiguration> obsconf;
  conf.get("covariancesObsAux", obsconf);
  ASSERT(covs_.size() == xx.size());
  for (std::size_t jobs = 0; jobs < xx.size(); ++jobs) {
    covs_[jobs]->update(xx[jobs], obsconf[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::update done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariances<MODEL>::linearize(const ObsAuxControls_ & xx,
                                         const ObsSpaces_ & odb) {
  Log::trace() << "ObsAuxCovariances<MODEL>::linearize starting" << std::endl;
  odb_ = &odb;
  ASSERT(covs_.size() == xx.size());
  for (std::size_t jobs = 0; jobs < xx.size(); ++jobs) {
    covs_[jobs]->linearize(xx[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::linearize done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariances<MODEL>::multiply(const ObsAuxIncrements_ & dx1,
                                       ObsAuxIncrements_ & dx2) const {
  Log::trace() << "ObsAuxCovariances<MODEL>::multiply starting" << std::endl;
  ASSERT(dx1.size() == dx2.size() && covs_.size() == dx1.size());
  for (std::size_t jobs = 0; jobs < dx1.size(); ++jobs) {
    covs_[jobs]->multiply(dx1[jobs], dx2[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::multiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariances<MODEL>::inverseMultiply(const ObsAuxIncrements_ & dx1,
                                              ObsAuxIncrements_ & dx2) const {
  Log::trace() << "ObsAuxCovariances<MODEL>::inverseMultiply starting" << std::endl;
  ASSERT(dx1.size() == dx2.size() && covs_.size() == dx1.size());
  for (std::size_t jobs = 0; jobs < dx1.size(); ++jobs) {
    covs_[jobs]->inverseMultiply(dx1[jobs], dx2[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::inverseMultiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariances<MODEL>::multiplySqrt(const ObsAuxCtlVecs_ & dv,
                                            ObsAuxIncrements_ & dx) const {
  Log::trace() << "ObsAuxCovariances<MODEL>::multiplySqrt starting" << std::endl;
  ASSERT(dv.size() == dx.size() && covs_.size() == dx.size());
  for (std::size_t jobs = 0; jobs < dx.size(); ++jobs) {
    covs_[jobs]->multiplySqrt(dv[jobs], dx[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::multiplySqrt done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariances<MODEL>::multiplySqrtTrans(const ObsAuxIncrements_ & dx,
                                                 ObsAuxCtlVecs_ & dv) const {
  Log::trace() << "ObsAuxCovariances<MODEL>::multiplySqrtTrans starting" << std::endl;
  ASSERT(dv.size() == dx.size() && covs_.size() == dx.size());
  for (std::size_t jobs = 0; jobs < dx.size(); ++jobs) {
    covs_[jobs]->multiplySqrtTrans(dx[jobs], dv[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::multiplySqrtTrans done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariances<MODEL>::randomize(ObsAuxIncrements_ & dx) const {
  Log::trace() << "ObsAuxCovariances<MODEL>::randomize starting" << std::endl;
  ASSERT(covs_.size() == dx.size());
  for (std::size_t jobs = 0; jobs < dx.size(); ++jobs) {
    covs_[jobs]->randomize(dx[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::randomize done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariances<MODEL>::print(std::ostream & os) const {
  Log::trace() << "ObsAuxCovariances<MODEL>::print starting" << std::endl;
  for (std::size_t jobs = 0; jobs < covs_.size(); ++jobs) os << *covs_.at(jobs) << " ";
  Log::trace() << "ObsAuxCovariances<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSAUXCOVARIANCES_H_
