/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_OBSERVERAD_H_
#define OOPS_BASE_OBSERVERAD_H_

#include <map>
#include <memory>
#include <vector>

#include "oops/base/Departures.h"
#include "oops/base/ObsAuxIncrements.h"
#include "oops/base/Observations.h"
#include "oops/base/PostBaseAD.h"
#include "oops/interface/GeoVaLs.h"
#include "oops/interface/Locations.h"
#include "oops/interface/ObservationSpace.h"
#include "oops/interface/LinearObsOperator.h"

#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

/// Computes observation equivalent AD and AD to/from increments.

template <typename MODEL, typename INCR> class ObserverAD : public PostBaseAD<INCR> {
  typedef Departures<MODEL>          Departures_;
  typedef GeoVaLs<MODEL>             GeoVaLs_;
  typedef LinearObsOperators<MODEL>  LinearObsOperators_;
  typedef Locations<MODEL>           Locations_;
  typedef ObsAuxIncrements<MODEL>    ObsAuxIncrs_;
  typedef ObsSpaces<MODEL>           ObsSpaces_;

 public:
  ObserverAD(const ObsSpaces_ &, const LinearObsOperators_ &,
             std::shared_ptr<const Departures_>, ObsAuxIncrs_ &,
             const util::Duration &, const bool subwin = false);
  ~ObserverAD() {}

  void setupAD(std::shared_ptr<const Departures_>, ObsAuxIncrs_ &);

 private:
// Methods
  void doFirstAD(INCR &, const util::DateTime &, const util::Duration &) override;
  void doProcessingAD(INCR &) override;
  void doLastAD(INCR &) override;

// Obs operator
  const ObsSpaces_ & obspaces_;
  const LinearObsOperators_ & hoptlad_;

// Data
  std::shared_ptr<const Departures_> ydep_;
  ObsAuxIncrs_ * dybias_;

  util::DateTime winbgn_;   //!< Begining of assimilation window
  util::DateTime winend_;   //!< End of assimilation window
  util::DateTime bgn_;      //!< Begining of currently active observations
  util::DateTime end_;      //!< End of currently active observations
  util::Duration hslot_;    //!< Half time slot
  const bool subwindows_;

  typedef typename std::map<util::DateTime, std::vector<std::unique_ptr<GeoVaLs_> > > gvalsMap_;
  typedef typename gvalsMap_::iterator gvalsIter_;
  gvalsMap_ gvals_;
};

// ====================================================================================

template <typename MODEL, typename INCR>
ObserverAD<MODEL, INCR>::ObserverAD(const ObsSpaces_ & obsdb,
                                    const LinearObsOperators_ & hoptlad,
                                    std::shared_ptr<const Departures_> ydep,
                                    ObsAuxIncrs_ & dybias,
                                    const util::Duration & tslot, const bool subwin)
  : PostBaseAD<INCR>(obsdb.windowStart(), obsdb.windowEnd()),
    obspaces_(obsdb), hoptlad_(hoptlad),
    ydep_(ydep), dybias_(&dybias),
    winbgn_(obsdb.windowStart()), winend_(obsdb.windowEnd()),
    bgn_(winbgn_), end_(winend_), hslot_(tslot/2), subwindows_(subwin)
{}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverAD<MODEL, INCR>::doFirstAD(INCR & dx, const util::DateTime & bgn,
                                        const util::Duration & tstep) {
  if (hslot_ == util::Duration(0)) hslot_ = tstep/2;
  const util::DateTime end(dx.validTime());
  if (subwindows_) {
    if (bgn == end) {
      bgn_ = bgn - hslot_;
      end_ = end + hslot_;
    } else {
      bgn_ = bgn;
      end_ = end;
    }
  }
  if (bgn_ < winbgn_) bgn_ = winbgn_;
  if (end_ > winend_) end_ = winend_;
  gvalsIter_ it = gvals_.find(bgn_);
  if (it == gvals_.end()) {
    std::vector<std::unique_ptr<GeoVaLs_> > gvalsvec;
    for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
      gvalsvec.emplace_back(new GeoVaLs_(obspaces_[jj], hoptlad_.variables(jj),
                                         dx, bgn_, end_));
    }
    gvals_.emplace(bgn_, std::move(gvalsvec));
    it = gvals_.find(bgn_);
    ASSERT(it != gvals_.end());
  }
  for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
    // [H(dx) P(dB)]^T
    hoptlad_[jj].obsEquivAD(*it->second[jj],
                            (*ydep_)[jj].obsvector(),
                            (*dybias_)[jj]);
    // P(dB)^T
    if ((*ydep_)[jj].hasBias())
      hoptlad_[jj].obsBiasEquivAD(*it->second[jj],
                                  (*ydep_)[jj].biasvector(),
                                  (*dybias_)[jj]);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverAD<MODEL, INCR>::doProcessingAD(INCR & dx) {
  util::DateTime t1(dx.validTime()-hslot_);
  util::DateTime t2(dx.validTime()+hslot_);
  if (t1 < bgn_) t1 = bgn_;
  if (t2 > end_) t2 = end_;

  for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
//  Get locations info for interpolator
    Locations_ locs(obspaces_[jj], t1, t2);

// Interpolate state variables to obs locations
    gvalsIter_ it = gvals_.find(bgn_);
    ASSERT(it != gvals_.end());
    dx.interpolateAD(locs, *it->second[jj]);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverAD<MODEL, INCR>::doLastAD(INCR &) {
  gvalsIter_ it = gvals_.find(bgn_);
  ASSERT(it != gvals_.end());
  for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
    it->second[jj]->zero();
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverAD<MODEL, INCR>::setupAD(std::shared_ptr<const Departures_> ydep,
                                      ObsAuxIncrs_ & dybias) {
  ydep_.reset();
  ydep_ = ydep;
  dybias_ = &dybias;
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSERVERAD_H_
