/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_GENERIC_DUALCONFIGMODEL_H_
#define OOPS_GENERIC_DUALCONFIGMODEL_H_

#include <string>

#include "eckit/memory/NonCopyable.h"

#include "oops/generic/DualConfigGeometry.h"
#include "oops/generic/DualConfigModelAuxControl.h"
#include "oops/generic/DualConfigPostProcessor.h"
#include "oops/generic/DualConfigState.h"

#include "oops/interface/Geometry.h"
#include "oops/interface/Model.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/State.h"

#include "util/Duration.h"
#include "util/LogbookWriter.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
  class Configuration;
}

namespace oops {

/// Encapsulates the nonlinear dual-config forecast model

// -----------------------------------------------------------------------------

template <typename MODEL>
class DualConfigModel : public util::Printable,
                        private eckit::NonCopyable,
                        private util::ObjectCounter<DualConfigModel<MODEL> >  {
  typedef DualConfigGeometry<MODEL>             DualConfigGeometry_;
  typedef DualConfigModel<MODEL>                DualConfigModel_;
  typedef DualConfigModelAuxControl<MODEL>      DualConfigModelAuxControl_;
  typedef DualConfigPostProcessor<MODEL>        DualConfigPostProcessor_;
  typedef DualConfigState<MODEL>                DualConfigState_;
  typedef Geometry<MODEL>                       Geometry_;
  typedef ModelAuxControl<MODEL>                ModelAux_;
  typedef Model<MODEL>                          Model_;
  typedef State<MODEL>                          State_;

 public:
  static const std::string classname() {return "oops::DualConfigModel<MODEL>";}

  DualConfigModel(const DualConfigGeometry_ &, const eckit::Configuration &);
  ~DualConfigModel();

// Run the forecast
  void forecast(DualConfigState_ &,
                const DualConfigModelAuxControl_ &,
                const util::Duration &,
                DualConfigPostProcessor<State_> &) const;

// Information and diagnostics
  const util::Duration & timeResolution1() const {return model1_->timeResolution();}
  const util::Duration & timeResolution2() const {return model2_->timeResolution();}

// Interfacing
  const Model_ & model1() const {return *model1_;}
  const Model_ & model2() const {return *model2_;}

 private:
  void initialize(DualConfigState_ &, const DualConfigModelAuxControl_ &) const;
  void step(DualConfigState_ &, const DualConfigModelAuxControl_ &) const;
  void finalize(DualConfigState_ &) const;
  void print(std::ostream &) const;

  std::unique_ptr<Model_> model1_;
  std::unique_ptr<Model_> model2_;
};

// =============================================================================

template<typename MODEL>
DualConfigModel<MODEL>::DualConfigModel(const DualConfigGeometry_ & resol,
                                        const eckit::Configuration & conf)
   : model1_(), model2_()
{
  Log::trace() << classname() << "::DualConfigModel starting" << std::endl;
  util::Timer timer(classname(), "DualConfigModel");
  Log::debug() << "Full config is:" << conf << std::endl;
  eckit::LocalConfiguration conf1(conf, "model1");
  model1_.reset(new Model_(resol.geometry1(), conf1));
  eckit::LocalConfiguration conf2(conf, "model2");
  model2_.reset(new Model_(resol.geometry2(), conf2));
  Log::trace() << classname() << "::DualConfigModel done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
DualConfigModel<MODEL>::~DualConfigModel() {
  Log::trace() << classname() << "::~DualConfigModel starting" << std::endl;
  util::Timer timer(classname(), "~DualConfigModel");
  model1_.reset();
  model2_.reset();
  Log::trace() << classname() << "::~DualConfigModel done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void DualConfigModel<MODEL>::forecast(DualConfigState_ & xx,
                                      const DualConfigModelAuxControl_ & maux,
                                      const util::Duration & len,
                                      DualConfigPostProcessor<State_> & post) const {
  Log::trace() << classname() << "::forecast starting" << std::endl;
  util::Timer timer(classname(), "forecast");
  util::LogbookWriter<std::string> log("Forecast","Nonlinear HR trajectory");

  const util::DateTime end(xx.validTime1() + len);
  Log::info() << classname() << "::forecast: starting 1: " << xx.state1() << std::endl;
  Log::info() << classname() << "::forecast: starting 2: " << xx.state2() << std::endl;

  this->initialize(xx, maux);
  post.post1().initialize(xx.state1(), end, model1_->timeResolution());
  post.post2().initialize(xx.state2(), end, model2_->timeResolution());
  while (xx.validTime1() < end) {
    Log::info() << classname() << "::step starting" << std::endl
                << "  t1 = " << xx.validTime1() << std::endl
                << "  t2 = " << xx.validTime2() << std::endl;
    if (xx.validTime2() == xx.validTime1()) xx.state2().forceWith(xx.state1());
    this->step(xx, maux);
    post.post1().process(xx.state1());
    if (xx.validTime2() == xx.validTime1()) post.post2().process(xx.state2());
    Log::info() << classname() << "::step done" << std::endl;
  }
  post.post1().finalize(xx.state1());
  post.post2().finalize(xx.state2());
  this->finalize(xx);
  Log::info() << classname() << "::forecast: forecast finished: " << xx << std::endl;
  ASSERT(xx.validTime1() == end);
  ASSERT(xx.validTime2() == end);

  Log::trace() << classname() << "::forecast done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void DualConfigModel<MODEL>::initialize(DualConfigState_ & xx,
                                        const DualConfigModelAuxControl_ & maux) const {
  Log::trace() << classname() << "initialize starting" << std::endl;
  util::Timer timer(classname(), "initialize");
  model1_->initialize(xx.state1(), maux.modelauxcontrol1());
  model2_->initialize(xx.state2(), maux.modelauxcontrol2());
  Log::trace() << classname() << "initialize done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void DualConfigModel<MODEL>::step(DualConfigState_ & xx,
                                  const DualConfigModelAuxControl_ & maux) const {
  Log::trace() << classname() << "::step starting" << std::endl;
  util::Timer timer(classname(), "step");
  model1_->step(xx.state1(), maux.modelauxcontrol1());
  if (xx.validTime2() < xx.validTime1())
    model2_->step(xx.state2(), maux.modelauxcontrol2());
  Log::trace() << classname() << "::step done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void DualConfigModel<MODEL>::finalize(DualConfigState_ & xx) const {
  Log::trace() << classname() << "::finalize starting" << std::endl;
  util::Timer timer(classname(), "finalize");
  model1_->finalize(xx.state1());
  model2_->finalize(xx.state2());
  Log::trace() << classname() << "::finalize done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void DualConfigModel<MODEL>::print(std::ostream & os) const {
  Log::trace() << classname() << "::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *model1_;
  os << *model2_;
  Log::trace() << classname() << "::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_GENERIC_DUALCONFIGMODEL_H_
