/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_GENERIC_DUALCONFIGSTATE_H_
#define OOPS_GENERIC_DUALCONFIGSTATE_H_

#include <memory> // for std::unique_ptr
#include <string>

#include "eckit/config/Configuration.h"

#include "oops/base/PostProcessor.h"
#include "oops/generic/DualConfigGeometry.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"

#include "util/DateTime.h"
#include "util/LogbookWriter.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {

template <typename MODEL>
class DualConfigModel;

template <typename MODEL>
class Model;

/// Encapsulates the model state

// -----------------------------------------------------------------------------

template <typename MODEL>
class DualConfigState : public util::Printable,
                        private util::ObjectCounter<DualConfigState<MODEL> > {
  typedef DualConfigGeometry<MODEL>  DualConfigGeometry_;
  typedef DualConfigModel<MODEL>     DualConfigModel_;
  typedef DualConfigState<MODEL>     DualConfigState_;
  typedef Geometry<MODEL>            Geometry_;
  typedef Model<MODEL>               Model_;
  typedef State<MODEL>               State_;

 public:
  static const std::string classname() {return "oops::DualConfigState";}

/// Constructor, destructor
  DualConfigState(const DualConfigGeometry_ &, const DualConfigModel_ &,
                  const eckit::Configuration &);
  DualConfigState(const DualConfigGeometry_ &, const DualConfigState &);
  DualConfigState(const DualConfigState &);
  ~DualConfigState();
  DualConfigState & operator=(const DualConfigState &);

/// Interfacing
  State_ & state1() {return *state1_;}
  State_ & state2() {return *state2_;}
  const State_ & state1() const {return *state1_;}
  const State_ & state2() const {return *state2_;}

/// Time
  const util::DateTime validTime1() const {return state1_->validTime();}
  const util::DateTime validTime2() const {return state2_->validTime();}

/// I/O and diagnostics
  void read(const eckit::Configuration &);
  void write(const eckit::Configuration &) const;
  double norm() const;

  Geometry_ geometry1() const;
  Geometry_ geometry2() const;

 protected:
  void zero();

 private:
  void print(std::ostream &) const;

  std::unique_ptr<State_> state1_;
  std::unique_ptr<State_> state2_;
};

// =============================================================================

template<typename MODEL>
DualConfigState<MODEL>::DualConfigState(const DualConfigGeometry_ & resol,
                                        const DualConfigModel_ & model,
                                        const eckit::Configuration & conf)
  : state1_(), state2_()
{
  Log::trace() << "DualConfigState<MODEL>::DualConfigState read starting" << std::endl;
  util::Timer timer(classname(), "DualConfigState");
  Log::debug() << "Full config is:" << conf << std::endl;
  eckit::LocalConfiguration conf1(conf, "state1");
  state1_.reset(new State_(resol.geometry1(), model.model1(), conf1));
  eckit::LocalConfiguration conf2(conf, "state2");
  state2_.reset(new State_(resol.geometry2(), model.model2(), conf2));
  Log::trace() << "DualConfigState<MODEL>::DualConfigState read done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
DualConfigState<MODEL>::DualConfigState(const DualConfigGeometry_ & resol,
                                        const DualConfigState & other)
  : state1_(), state2_()
{
  Log::trace() << "DualConfigState<MODEL>::DualConfigState interpolated starting" << std::endl;
  util::Timer timer(classname(), "DualConfigState");
  state1_.reset(new State_(resol.geometry1(), *other.state1_));
  state2_.reset(new State_(resol.geometry2(), *other.state2_));
  Log::trace() << "DualConfigState<MODEL>::DualConfigState interpolated done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
DualConfigState<MODEL>::DualConfigState(const DualConfigState & other)
  : state1_(), state2_()
{
  Log::trace() << "DualConfigState<MODEL>::DualConfigState starting copy" << std::endl;
  util::Timer timer(classname(), "DualConfigState");
  state1_.reset(new State_(*other.state1_));
  state2_.reset(new State_(*other.state2_));
  Log::trace() << "DualConfigState<MODEL>::DualConfigState copy done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
DualConfigState<MODEL>::~DualConfigState() {
  Log::trace() << "DualConfigState<MODEL>::~DualConfigState starting" << std::endl;
  util::Timer timer(classname(), "~DualConfigState");
  state1_.reset();
  state2_.reset();
  Log::trace() << "DualConfigState<MODEL>::~DualConfigState done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
DualConfigState<MODEL> & DualConfigState<MODEL>::operator=(const DualConfigState & rhs) {
  Log::trace() << "DualConfigState<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");
  *state1_ = *rhs.state1_;
  *state2_ = *rhs.state2_;
  Log::trace() << "DualConfigState<MODEL>::operator= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void DualConfigState<MODEL>::read(const eckit::Configuration & conf) {
  Log::trace() << "DualConfigState<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");
  Log::debug() << "Full config is:" << conf << std::endl;
  eckit::LocalConfiguration conf1(conf, "model1");
  state1_->read(conf1);
  eckit::LocalConfiguration conf2(conf, "model2");
  state2_->read(conf2);
  Log::trace() << "DualConfigState<MODEL>::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void DualConfigState<MODEL>::write(const eckit::Configuration & conf) const {
  Log::trace() << "DualConfigState<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");
  util::LogbookWriter<bool> log("DualConfigStateWriting", true);
  Log::debug() << "Full config is:" << conf << std::endl;
  eckit::LocalConfiguration conf1(conf, "model1");
  state1_->write(conf1);
  eckit::LocalConfiguration conf2(conf, "model2");
  state2_->write(conf2);
  Log::trace() << "DualConfigState<MODEL>::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
double DualConfigState<MODEL>::norm() const {
  Log::trace() << "DualConfigState<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = state1_->norm();
  zz += state2_->norm();
  Log::trace() << "DualConfigState<MODEL>::norm done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
Geometry<MODEL> DualConfigState<MODEL>::geometry1() const {
  Log::trace() << "DualConfigState<MODEL>::geometry1 starting" << std::endl;
  util::Timer timer(classname(), "geometry1");
  Geometry<MODEL> geom(state1_->geometry());
  Log::trace() << "DualConfigState<MODEL>::geometry1 done" << std::endl;
  return geom;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
Geometry<MODEL> DualConfigState<MODEL>::geometry2() const {
  Log::trace() << "DualConfigState<MODEL>::geometry2 starting" << std::endl;
  util::Timer timer(classname(), "geometry2");
  Geometry<MODEL> geom(state2_->geometry());
  Log::trace() << "DualConfigState<MODEL>::geometry2 done" << std::endl;
  return geom;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void DualConfigState<MODEL>::print(std::ostream & os) const {
  Log::trace() << "DualConfigState<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *state1_;
  os << *state2_;
  Log::trace() << "DualConfigState<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void DualConfigState<MODEL>::zero() {
  Log::trace() << "DualConfigState<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  state1_->zero();
  state2_->zero();
  Log::trace() << "DualConfigState<MODEL>::zero done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_GENERIC_DUALCONFIGSTATE_H_
