/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_GENERIC_DUALCONFIGVARIABLES_H_
#define OOPS_GENERIC_DUALCONFIGVARIABLES_H_

#include <memory>
#include <ostream>
#include <string>

#include "interface/Variables.h"

#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
  class Configuration;
}

namespace oops {

// -----------------------------------------------------------------------------

template <typename MODEL>
class DualConfigVariables : public util::Printable,
                            private util::ObjectCounter<DualConfigVariables<MODEL> > {
  typedef Variables<MODEL> Variables_

 public:
  static const std::string classname() {return "oops::DualConfigVariables";}

  explicit DualConfigVariables(const eckit::Configuration &);
  DualConfigVariables(const DualConfigVariables &);
  ~DualConfigVariables();

/// Interfacing
  const Variables_ & variables1() const {return *variables1_;}
  const Variables_ & variables2() const {return *variables2_;}

 private:
  DualConfigVariables & operator=(const DualConfigVariables &);
  void print(std::ostream &) const;

  std::shared_ptr<const Variables_> variables1_;
  std::shared_ptr<const Variables_> variables2_;
};

// -----------------------------------------------------------------------------

template <typename MODEL>
DualConfigVariables<MODEL>::DualConfigVariables(const eckit::Configuration & conf)
  : variables1_(), variables2_()
{
  Log::trace() << "DualConfigVariables<MODEL>::DualConfigVariables starting" << std::endl;
  util::Timer timer(classname(), "DualConfigVariables");
  Log::debug() << "Full config is:" << conf << std::endl;
  eckit::LocalConfiguration conf1(conf, "model1");
  variables1_.reset(new Variables_(conf1));
  eckit::LocalConfiguration conf1(conf, "model2");
  variables1_.reset(new Variables_(conf2));
  Log::trace() << "DualConfigVariables<MODEL>::DualConfigVariables done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
DualConfigVariables<MODEL>::DualConfigVariables(const DualConfigVariables & other) {
  Log::trace() << "DualConfigVariables<MODEL>::DualConfigVariables copy starting" << std::endl;
  util::Timer timer(classname(), "DualConfigVariables");
  variables1_.reset(other.variables1_);
  variables2_.reset(other.variables2_);
  Log::trace() << "DualConfigVariables<MODEL>::DualConfigVariables copy done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
DualConfigVariables<MODEL>::~DualConfigVariables() {
  Log::trace() << "DualConfigVariables<MODEL>::~DualConfigVariables starting" << std::endl;
  util::Timer timer(classname(), "~DualConfigVariables");
  variables1_.reset();
  variables2_.reset();
  Log::trace() << "DualConfigVariables<MODEL>::~DualConfigVariables done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void DualConfigVariables<MODEL>::print(std::ostream & os) const {
  Log::trace() << "DualConfigVariables<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *variables1_;
  os << *variables2_;
  Log::trace() << "DualConfigVariables<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_GENERIC_DUALCONFIGVARIABLES_H_
