/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_JCCOVARBASE_H_
#define OOPS_INTERFACE_JCCOVARBASE_H_

#include <map>
#include <memory>
#include <string>

#include "eckit/memory/NonCopyable.h"

#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"

#include "util/abor1_cpp.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
  class Configuration;
}

namespace oops {

/// Base class for encapsulation of the Jc covariances.
/*!
 * Defines the interfaces for Jc covariances.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class JcCovarBase : public util::Printable,
                    private eckit::NonCopyable {
  typedef Increment<MODEL>              Increment_;
  typedef Geometry<MODEL>               Geometry_;
  typedef State<MODEL>                  State_;

 public:
  static const std::string classname() {return "oops::JcCovarBase";}

/// Constructors, destructors
  JcCovarBase() {}
  virtual ~JcCovarBase() {}

/// Methods
  void linearize(const State_ &, const Geometry_ &);
  void multiply(const Increment_ &, Increment_ &) const;
  void inverseMultiply(const Increment_ &, Increment_ &) const;

 protected:
  virtual void linearize(const typename MODEL::State &, const typename MODEL::Geometry &) = 0;
  virtual void multiply(const typename MODEL::Increment &, typename MODEL::Increment &) const = 0;
  virtual void inverseMultiply(const typename MODEL::Increment &, typename MODEL::Increment &) const = 0;

/// Information and diagnostics
  virtual void print(std::ostream &) const = 0;
};

// =============================================================================

/// JcCovarFactory Factory
template <typename MODEL>
class JcCovarFactory {
  typedef Geometry<MODEL>   Geometry_;
  typedef State<MODEL>      State_;
 public:
  static JcCovarBase<MODEL> * create(const State_ &, const Geometry_ &,
                                     const eckit::Configuration &);
  virtual ~JcCovarFactory() { getMakers().clear(); }
 protected:
  explicit JcCovarFactory(const std::string &);
 private:
  virtual JcCovarBase<MODEL> * make(const State_ &, const Geometry_ &,
                                    const eckit::Configuration &) =0;
  static std::map < std::string, JcCovarFactory<MODEL> * > & getMakers() {
    static std::map < std::string, JcCovarFactory<MODEL> * > makers_;
    return makers_;
  }
};

// -----------------------------------------------------------------------------

template<class MODEL, class T>
class JcCovarMaker : public JcCovarFactory<MODEL> {
  typedef Geometry<MODEL>   Geometry_;
  typedef State<MODEL>      State_;
  virtual JcCovarBase<MODEL> * make(const State_ & state, const Geometry_ & geom,
                                    const eckit::Configuration & conf)
    { return new T(state.state(), geom.geometry(), conf); }
 public:
  explicit JcCovarMaker(const std::string & name) : JcCovarFactory<MODEL>(name) {}
};

// -----------------------------------------------------------------------------

template <typename MODEL>
JcCovarFactory<MODEL>::JcCovarFactory(const std::string & name) {
  if (getMakers().find(name) != getMakers().end()) {
    Log::error() << name << " already registered in the Jc covariances factory." << std::endl;
    ABORT("Element already registered in JcCovarFactory.");
  }
  getMakers()[name] = this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
JcCovarBase<MODEL>* JcCovarFactory<MODEL>::create(const State_ & state, 
                                                  const Geometry_ & geom,
                                                  const eckit::Configuration & conf) {
  Log::trace() << "JcCovarBase<MODEL>::create starting" << std::endl;
  const std::string id = conf.getString("version");
  typename std::map<std::string, JcCovarFactory<MODEL>*>::iterator
    jerr = getMakers().find(id);
  if (jerr == getMakers().end()) {
    Log::error() << id << " does not exist in the Jc covariances factory." << std::endl;
    ABORT("Element does not exist in JcCovarFactory.");
  }
  JcCovarBase<MODEL> * ptr = jerr->second->make(state, geom, conf);
  Log::trace() << "JcCovarBase<MODEL>::create done" << std::endl;
  return ptr;
}

// =============================================================================

template<typename MODEL>
void JcCovarBase<MODEL>::linearize(const State_ & x,
                                   const Geometry_ & geom) {
  Log::trace() << "JcCovarBase<MODEL>::linearize starting" << std::endl;
  util::Timer timer(classname(), "linearize");
  this->linearize(x.state(), geom.geometry());
  Log::trace() << "JcCovarBase<MODEL>::linearize done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void JcCovarBase<MODEL>::multiply(const Increment_ & dx_in,
                                  Increment_ & dx_out) const {
  Log::trace() << "JcCovarBase<MODEL>::multiply starting" << std::endl;
  util::Timer timer(classname(), "multiply");
  this->multiply(dx_in.increment(), dx_out.increment());
  Log::trace() << "JcCovarBase<MODEL>::multiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void JcCovarBase<MODEL>::inverseMultiply(const Increment_ & dx_in,
                                         Increment_ & dx_out) const {
  Log::trace() << "JcCovarBase<MODEL>::inverseMultiply starting" << std::endl;
  util::Timer timer(classname(), "inverseMultiply");
  this->inverseMultiply(dx_in.increment(), dx_out.increment());
  Log::trace() << "JcCovarBase<MODEL>::inverseMultiply done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_JCCOVARBASE_H_
