/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSAUXCONTROL_H_
#define OOPS_INTERFACE_OBSAUXCONTROL_H_

#include <cmath>
#include <iostream>
#include <map>
#include <memory>
#include <string>

#include "eckit/exception/Exceptions.h"

#include "oops/interface/ObsAuxControlBase.h"
#include "oops/interface/ObservationSpace.h"

#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
  class Configuration;
}

namespace oops {

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxControl : public util::Printable,
                      private util::ObjectCounter<ObsAuxControl<MODEL> > {
  typedef ObsAuxControlBase<MODEL> ObsAuxControlBase_;
  typedef ObservationSpace<MODEL>  ObservationSpace_;

  typedef std::map<std::string, std::unique_ptr<ObsAuxControlBase_> >
    ObsAuxControlPtrMap_;
  typedef typename ObsAuxControlPtrMap_::iterator iter_;
  typedef typename ObsAuxControlPtrMap_::const_iterator icst_;

 public:
  static const std::string classname() {return "oops::ObsAuxControl";}

/// Constructor, destructor
  ObsAuxControl(const ObservationSpace_ &, const eckit::Configuration &);
  explicit ObsAuxControl(const ObsAuxControl &, const bool copy = true);
  ~ObsAuxControl();

/// Reset
  void reset(const ObservationSpace_ &, const std::string &, const eckit::Configuration &);

/// I/O and diagnostics
  void read(const eckit::Configuration &);
  void write(const eckit::Configuration &) const;
  double norm() const;

/// Interfacing
  const ObsAuxControlBase_ & obsauxcontrol(const std::string & id) const {
    return this->get(id);
  }
  ObsAuxControlBase_ & obsauxcontrol(const std::string & id) {
    return this->get(id);
  }

  const ObsAuxControlPtrMap_ & obsauxcontrol() const {return ctls_;}
  ObsAuxControlPtrMap_ & obsauxcontrol() {return ctls_;}

 private:
  ObsAuxControl & operator=(const ObsAuxControl &);
  ObsAuxControlBase_ & get(const std::string &);
  const ObsAuxControlBase_ & get(const std::string &) const;
  void print(std::ostream &) const;

  mutable size_t size_;
  ObsAuxControlPtrMap_ ctls_;
};

// =============================================================================

template<typename MODEL>
ObsAuxControl<MODEL>::ObsAuxControl(const ObservationSpace_ & os,
                                    const eckit::Configuration & conf)
  : size_(0), ctls_()
{
  Log::trace() << "ObsAuxControl<MODEL>::ObsAuxControl starting" << std::endl;
  util::Timer timer(classname(), "ObsAuxControl");

  std::vector<eckit::LocalConfiguration> confs;
  if (conf.has("stateObsAux")) conf.get("stateObsAux", confs);
  size_ = confs.size();

  for (auto const & config: confs) {
    std::string key(config.getString("type"));
    std::unique_ptr<ObsAuxControlBase<MODEL>>
      auxctl(ObsAuxControlFactory<MODEL>::create(os, config));
    ctls_.emplace(key, std::move(auxctl));
  }
  Log::trace() << "ObsAuxControl<MODEL>::ObsAuxControl done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
ObsAuxControl<MODEL>::ObsAuxControl(const ObsAuxControl & other, const bool copy)
  : size_(other.size_), ctls_()
{
  Log::trace() << "ObsAuxControl<MODEL>::ObsAuxControl copy starting" << std::endl;
  util::Timer timer(classname(), "ObsAuxControl");
  for (icst_ it = other.ctls_.begin(); it != other.ctls_.end(); ++it) {
    std::string incname = it->first;
    std::unique_ptr<ObsAuxControlBase<MODEL>> auxctl(it->second->clone(copy));
    ctls_.emplace(incname, std::move(auxctl));
  }
  Log::trace() << "ObsAuxControl<MODEL>::ObsAuxControl copy done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
ObsAuxControl<MODEL>::~ObsAuxControl() {
  Log::trace() << "ObsAuxControl<MODEL>::~ObsAuxControl starting" << std::endl;
  util::Timer timer(classname(), "~ObsAuxControl");
  Log::trace() << "ObsAuxControl<MODEL>::~ObsAuxControl done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
ObsAuxControlBase<MODEL> & ObsAuxControl<MODEL>::get(const std::string & id) {
  iter_ it = ctls_.find(id);
  ASSERT(it != ctls_.end());
  return *it->second;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
const ObsAuxControlBase<MODEL> & ObsAuxControl<MODEL>::get(const std::string & id) const {
  icst_ it = ctls_.find(id);
  ASSERT(it != ctls_.end());
  return *it->second;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxControl<MODEL>::reset(const ObservationSpace_ & os,
                                 const std::string & stateid,
                                 const eckit::Configuration & conf) {
  Log::trace() << "ObsAuxControl<MODEL>::reset starting" << std::endl;
  util::Timer timer(classname(), "reset");

  std::vector<eckit::LocalConfiguration> confs;
  if (conf.has("stateObsAuxReset")) {
    conf.get("stateObsAuxReset", confs);
  }

  for (iter_ it = ctls_.begin(); it != ctls_.end(); ++it) {
    for (auto const & config: confs) {
      if (config.getString("type") == it->first)
        it->second->reset(os, stateid, config);
    }
  }
  Log::trace() << "ObsAuxControl<MODEL>::reset done" << std::endl;
}


// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxControl<MODEL>::read(const eckit::Configuration & conf) {
  Log::trace() << "ObsAuxControl<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");

  std::vector<eckit::LocalConfiguration> confs;
  if (conf.has("stateObsAux")) conf.get("stateObsAux", confs);

  for (iter_ it = ctls_.begin(); it != ctls_.end(); ++it) {
    for (auto const & config: confs) {
      if (config.getString("type") == it->first)
        it->second->read(config);
    }
  }
  Log::trace() << "ObsAuxControl<MODEL>::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxControl<MODEL>::write(const eckit::Configuration & conf) const {
  Log::trace() << "ObsAuxControl<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");

  std::vector<eckit::LocalConfiguration> confs;
  if (conf.has("stateObsAux")) conf.get("stateObsAux", confs);

  for (icst_ it = ctls_.begin(); it != ctls_.end(); ++it) {
    for (auto const & config: confs) {
      if (config.getString("type") == it->first)
        it->second->write(config);
    }
  }
  Log::trace() << "ObsAuxControl<MODEL>::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
double ObsAuxControl<MODEL>::norm() const {
  Log::trace() << "ObsAuxControl<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = 0.0;
  for (icst_ it = ctls_.begin(); it != ctls_.end(); ++it) {
    double norm = it->second->norm();
    zz += norm*norm;
  }
  Log::trace() << "ObsAuxControl<MODEL>::norm done" << std::endl;
  return std::sqrt(zz);
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxControl<MODEL>::print(std::ostream & os) const {
  Log::trace() << "ObsAuxControl<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  for (icst_ it = ctls_.begin(); it != ctls_.end(); ++it) {
    os << it->first << ": " << *it->second;
  }
  Log::trace() << "ObsAuxControl<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSAUXCONTROL_H_
