/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSAUXCOVARIANCEBASE_H_
#define OOPS_INTERFACE_OBSAUXCOVARIANCEBASE_H_

#include <iostream>
#include <memory> // for std::unique_ptr
#include <string>

#include "eckit/config/Configuration.h"
#include "eckit/memory/NonCopyable.h"

#include "oops/interface/ObservationSpace.h"

#include "util/abor1_cpp.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {

template <typename MODEL>
class ObsAuxControlBase;

template <typename MODEL>
class ObsAuxCtlVecBase;

template <typename MODEL>
class ObsAuxIncrementBase;

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxCovarianceBase : public util::Printable,
                             private eckit::NonCopyable {
  typedef ObsAuxControlBase<MODEL>   ObsAuxControlBase_;
  typedef ObsAuxCtlVecBase<MODEL>    ObsAuxCtlVecBase_;
  typedef ObsAuxIncrementBase<MODEL> ObsAuxIncrementBase_;

 public:
  static const std::string classname() {return "oops::ObsAuxCovariance";}

  ObsAuxCovarianceBase() {}
  virtual ~ObsAuxCovarianceBase() {}

/// Update
  virtual void update(const ObsAuxControlBase_ &, const eckit::Configuration &) =0;

/// Operators
  virtual void linearize(const ObsAuxControlBase_ &) =0;
  virtual void multiply(const ObsAuxIncrementBase_ &, ObsAuxIncrementBase_ &) const =0;
  virtual void inverseMultiply(const ObsAuxIncrementBase_ &, ObsAuxIncrementBase_ &) const =0;
  virtual void multiplySqrt(const ObsAuxCtlVecBase_ &, ObsAuxIncrementBase_ &) const =0;
  virtual void multiplySqrtTrans(const ObsAuxIncrementBase_ &, ObsAuxCtlVecBase_ &) const =0;
  virtual void randomize(ObsAuxIncrementBase_ &) const =0;

// Interfacing
  virtual const eckit::Configuration & config() const =0;
  virtual const int & getAccess() const =0;
  virtual int & getAccess() =0;

 private:
  virtual void print(std::ostream &) const =0;
};

// =============================================================================

/// ObsAuxCovariance Factory
template <typename MODEL>
class ObsAuxCovarianceFactory {
 public:
  static ObsAuxCovarianceBase<MODEL> * create(const ObservationSpace<MODEL> &,
                                              const eckit::Configuration &);
  virtual ~ObsAuxCovarianceFactory() { getMakers().clear(); }
 protected:
  explicit ObsAuxCovarianceFactory(const std::string &);
 private:
  virtual ObsAuxCovarianceBase<MODEL> * make(const ObservationSpace<MODEL> &,
                                            const eckit::Configuration &) =0;
  static std::map < std::string, ObsAuxCovarianceFactory<MODEL> * > & getMakers() {
    static std::map < std::string, ObsAuxCovarianceFactory<MODEL> * > makers_;
    return makers_;
  }
};

// -----------------------------------------------------------------------------

template<typename MODEL, class T>
class ObsAuxCovarianceMaker : public ObsAuxCovarianceFactory<MODEL> {
  virtual ObsAuxCovarianceBase<MODEL> * make(const ObservationSpace<MODEL> & os,
                                             const eckit::Configuration & conf)
    { return new T(os.observationspace(), conf); }
 public:
  explicit ObsAuxCovarianceMaker(const std::string & name)
    : ObsAuxCovarianceFactory<MODEL>(name) {}
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxCovarianceFactory<MODEL>::ObsAuxCovarianceFactory(const std::string & name) {
  if (getMakers().find(name) != getMakers().end()) {
    Log::error() << name << " already registered in ObsAuxCovarianceFactory."
                 << std::endl;
    ABORT("Element already registered in ObsAuxCovarianceFactory.");
  }
  getMakers()[name] = this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxCovarianceBase<MODEL>*
ObsAuxCovarianceFactory<MODEL>::create(const ObservationSpace<MODEL> & os,
                                       const eckit::Configuration & conf) {
  const std::string id = conf.getString("type");
  typename std::map<std::string, ObsAuxCovarianceFactory<MODEL>*>::iterator
    jloc = getMakers().find(id);
  if (jloc == getMakers().end()) {
    Log::error() << id << " does not exist in ObsAuxCovarianceFactory."
                 << std::endl;
    ABORT("Element does not exist in ObsAuxCovarianceFactory.");
  }
  ObsAuxCovarianceBase<MODEL> * ptr = jloc->second->make(os, conf);
  return ptr;
}

// =============================================================================

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSAUXCOVARIANCEBASE_H_
